mod msg;
mod node;

mod io;
mod service;

pub mod prelude {
    pub use crate::msg::Message;
}

use io::{FleckIO,IOChannel};

pub trait ServiceInterface {
    fn send(&self, packet: io::Packet);
}

pub trait FleckService {
    fn process_incoming(&self, packet: &mut io::Packet) {}
    fn process_outgoing(&self, packet: &mut io::Packet) {}
    fn process_minor_tick(&self, si: &dyn ServiceInterface) {}
    fn process_major_tick(&self, si: &dyn ServiceInterface) {}
}

pub struct Fleck {
    io: std::rc::Rc<dyn FleckIO>,
    services: Vec<std::rc::Rc<dyn FleckService>>,
}

impl Fleck {
    pub fn new() -> Self {
        let mut res = Self {
            io: io::platform(),
            services: Vec::new()
        };

        res.register_services();

        res
    }

    /*pub fn new_with_io(io: std::rc::Rc<dyn FleckIO>) -> Self {
        Self {
            io,
            services: Vec::new()
        }
    }*/

    fn register_services(&mut self) {
        // Inserted in processing order
        // Local discovery
        self.services.push(std::rc::Rc::new(service::LocalDiscovery::new(self.io.clone())));

        // Finally, send the packet
        self.services.push(std::rc::Rc::new(service::SendPacket::new(self.io.clone())));
    }

    pub fn register_service(&mut self, srv: std::rc::Rc<dyn FleckService>) {
        self.services.push(srv);
    }

    pub fn run(&self) {
        self.io.poll(self);
    }
}

impl ServiceInterface for Fleck {
     fn send(&self, mut packet: io::Packet) {
        // process services in forwards order
        for svc in self.services.iter() {
            svc.process_outgoing(&mut packet);
        }
     }
}

impl io::IOFeedback for Fleck {
    fn packet(&self, mut packet: io::Packet) {
        println!("received a packet!");
        // received a packet!
        // process services in reverse order
        for svc in self.services.iter().rev() {
            svc.process_incoming(&mut packet);
        }
    }

    fn minor_tick(&self) {
        // process services in forwards order
        for svc in self.services.iter() {
            svc.process_minor_tick(self);
        }
    }

    fn major_tick(&self) {
        // process services in forwards order
        for svc in self.services.iter() {
            svc.process_major_tick(self);
        }
    }
}
